@@ -10,7 +10,7 @@
   See the file COPYING.LIB.
 */
 
-#include "fuse_config.h"
+#include "config.h"
 #include "fuse_i.h"
 #include "fuse_misc.h"
 #include "fuse_opt.h"
@@ -50,7 +50,6 @@ static const struct fuse_opt fuse_helper_opts[] = {
 #endif
 	FUSE_HELPER_OPT("clone_fd",	clone_fd),
 	FUSE_HELPER_OPT("max_idle_threads=%u", max_idle_threads),
-	FUSE_HELPER_OPT("max_threads=%u", max_threads),
 	FUSE_OPT_END
 };
 
@@ -137,8 +136,6 @@ void fuse_cmdline_help(void)
 	       "    -o clone_fd            use separate fuse device fd for each thread\n"
 	       "                           (may improve performance)\n"
 	       "    -o max_idle_threads    the maximum number of idle worker threads\n"
-	       "                           allowed (default: -1)\n"
-	       "    -o max_threads         the maximum number of worker threads\n"
 	       "                           allowed (default: 10)\n");
 }
 
@@ -157,14 +154,14 @@ static int fuse_helper_opt_proc(void *data, const char *arg, int key,
 
 			char mountpoint[PATH_MAX] = "";
 			if (realpath(arg, mountpoint) == NULL) {
-				fuse_log(FUSE_LOG_ERR,
+				fprintf(stderr,
 					"fuse: bad mount point `%s': %s\n",
 					arg, strerror(errno));
 				return -1;
 			}
 			return fuse_opt_add_opt(&opts->mountpoint, mountpoint);
 		} else {
-			fuse_log(FUSE_LOG_ERR, "fuse: invalid argument `%s'\n", arg);
+			fprintf(stderr, "fuse: invalid argument `%s'\n", arg);
 			return -1;
 		}
 
@@ -189,7 +186,7 @@ static int add_default_subtype(const char *progname, struct fuse_args *args)
 
 	subtype_opt = (char *) malloc(strlen(basename) + 64);
 	if (subtype_opt == NULL) {
-		fuse_log(FUSE_LOG_ERR, "fuse: memory allocation failed\n");
+		fprintf(stderr, "fuse: memory allocation failed\n");
 		return -1;
 	}
 #ifdef __FreeBSD__
@@ -202,16 +199,12 @@ static int add_default_subtype(const char *progname, struct fuse_args *args)
 	return res;
 }
 
-int fuse_parse_cmdline_312(struct fuse_args *args,
-			   struct fuse_cmdline_opts *opts);
-FUSE_SYMVER("fuse_parse_cmdline_312", "fuse_parse_cmdline@@FUSE_3.12")
-int fuse_parse_cmdline_312(struct fuse_args *args,
-			   struct fuse_cmdline_opts *opts)
+int fuse_parse_cmdline(struct fuse_args *args,
+		       struct fuse_cmdline_opts *opts)
 {
 	memset(opts, 0, sizeof(struct fuse_cmdline_opts));
 
-	opts->max_idle_threads = -1; /* new default in fuse version 3.12 */
-	opts->max_threads = 10;
+	opts->max_idle_threads = 10;
 
 	if (fuse_opt_parse(args, opts, fuse_helper_opts,
 			   fuse_helper_opt_proc) == -1)
@@ -228,28 +221,6 @@ int fuse_parse_cmdline_312(struct fuse_args *args,
 	return 0;
 }
 
-/**
- * struct fuse_cmdline_opts got extended in libfuse-3.12
- */
-int fuse_parse_cmdline_30(struct fuse_args *args,
-		       struct fuse_cmdline_opts *opts);
-FUSE_SYMVER("fuse_parse_cmdline_30", "fuse_parse_cmdline@FUSE_3.0")
-int fuse_parse_cmdline_30(struct fuse_args *args,
-			  struct fuse_cmdline_opts *out_opts)
-{
-	struct fuse_cmdline_opts opts;
-
-
-	int rc = fuse_parse_cmdline_312(args, &opts);
-	if (rc == 0) {
-		/* copy up to the size of the old pre 3.12 struct */
-		memcpy(out_opts, &opts,
-		       offsetof(struct fuse_cmdline_opts, max_idle_threads) +
-		       sizeof(opts.max_idle_threads));
-	}
-
-	return rc;
-}
 
 int fuse_daemonize(int foreground)
 {
@@ -312,7 +283,6 @@ int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
 	struct fuse *fuse;
 	struct fuse_cmdline_opts opts;
 	int res;
-	struct fuse_loop_config *loop_config = NULL;
 
 	if (fuse_parse_cmdline(&args, &opts) != 0)
 		return 1;
@@ -337,7 +307,7 @@ int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
 
 	if (!opts.show_help &&
 	    !opts.mountpoint) {
-		fuse_log(FUSE_LOG_ERR, "error: no mountpoint specified\n");
+		fprintf(stderr, "error: no mountpoint specified\n");
 		res = 2;
 		goto out1;
 	}
@@ -368,19 +338,13 @@ int fuse_main_real(int argc, char *argv[], const struct fuse_operations *op,
 	if (opts.singlethread)
 		res = fuse_loop(fuse);
 	else {
-		loop_config = fuse_loop_cfg_create();
-		if (loop_config == NULL) {
-			res = 7;
-			goto out3;
-		}
-
-		fuse_loop_cfg_set_clone_fd(loop_config, opts.clone_fd);
-
-		fuse_loop_cfg_set_idle_threads(loop_config, opts.max_idle_threads);
-		res = fuse_loop_mt(fuse, loop_config);
+		struct fuse_loop_config loop_config;
+		loop_config.clone_fd = opts.clone_fd;
+		loop_config.max_idle_threads = opts.max_idle_threads;
+		res = fuse_loop_mt_32(fuse, &loop_config);
 	}
 	if (res)
-		res = 8;
+		res = 7;
 
 	fuse_remove_signal_handlers(se);
 out3:
@@ -388,7 +352,6 @@ out3:
 out2:
 	fuse_destroy(fuse);
 out1:
-	fuse_loop_cfg_destroy(loop_config);
 	free(opts.mountpoint);
 	fuse_opt_free_args(&args);
 	return res;
@@ -448,7 +411,7 @@ struct fuse_conn_info_opts* fuse_parse_conn_info_opts(struct fuse_args *args)
 
 	opts = calloc(1, sizeof(struct fuse_conn_info_opts));
 	if(opts == NULL) {
-		fuse_log(FUSE_LOG_ERR, "calloc failed\n");
+		fprintf(stderr, "calloc failed\n");
 		return NULL;
 	}
 	if(fuse_opt_parse(args, opts, conn_info_opt_spec, NULL) == -1) {
